unit GraphQL.Bookshelf;

interface

uses
  SysUtils, Generics.Collections, Rtti, TypInfo,
  Bcl.Utils,
  GraphQL.Document.Values,
  GraphQL.Executer,
  GraphQL.Schema;

type
  TBook = class;

  TAuthor = class
  private
    FId: Integer;
    FName: string;
    FCreatedOn: TDateTime;
    function GetBooks: TArray<TBook>;
  public
    constructor Create(AName: string);
    property Books: TArray<TBook> read GetBooks;

    property Id: Integer read FId;
    property Name: string read FName;
    property CreatedOn: TDateTime read FCreatedOn;
  end;

  TBook = class
  private
    FId: Integer;
    FName: string;
    FAuthorId: Integer;
    function GetAuthor: TAuthor;
  public
    constructor Create(AName: string; AAuthorID: Integer);
    property Id: Integer read FId;
    property Name: string read FName;
    property Author: TAuthor read GetAuthor;
  end;

  TTestInput = class
  private
    FMyName: String;
    FMyInput: String;
  public
    constructor Create(AName: String);
    property MyName: String read FMyName write FMyName;
    property MyInput: String read FMyInput write FMyInput;
  end;

  TTestOutput = class
  private
    FMyName: String;
    FMyOutput: String;
  public
    constructor Create(AName: String);
    property MyName: String read FMyName write FMyName;
    property MyOutput: String read FMyOutput write FMyOutput;
  end;

  TMutation = class
  public
    function CreateAuthor(Name: string): TAuthor;
    function CreateBook(Name: string; AuthorID: Integer): TBook;
    function TestUpdate(Input: TTestInput): String;
  end;

  TQuery = class
    function Books: TArray<TBook>;
    function Authors: TArray<TAuthor>;
    Function Book(id: Integer): TBook;
    function Author(id: Integer): TAuthor;
  end;

  TSchemaDateTimeType = class(TSchemaScalarType)
  strict protected
    function ParseLiteral(Value: TASTValue): TValue; override;
  public
    constructor Create; reintroduce;
  public
    function ParseValue(const Value: TValue): TValue; override;
    function Serialize(const Value: TValue): TValue; override;
  end;

const
  BookshelfSchema =
    'type Book {                                     ' + sLineBreak +
    '  id: ID!                                       ' + sLineBreak +
    '  name: String!                                 ' + sLineBreak +
    '  author: Author!                               ' + sLineBreak +
    '}                                               ' + sLineBreak +
    '                                                ' + sLineBreak +
    'type Author {                                   ' + sLineBreak +
    '  id: ID!                                       ' + sLineBreak +
    '  name: String!                                 ' + sLineBreak +
    '  createdOn: DateTime!                          ' + sLineBreak +
    '  books: [Book!]!                               ' + sLineBreak +
    '}                                               ' + sLineBreak +
    '                                                ' + sLineBreak +
    'input TestInput {                               ' + sLineBreak +
    '  myInput: String                               ' + sLineBreak +
    '}                                               ' + sLineBreak +
    '                                                ' + sLineBreak +
    'type TestOutput {                               ' + sLineBreak +
    '  myOuput: String                               ' + sLineBreak +
    '}                                               ' + sLineBreak +
    '                                                ' + sLineBreak +
    'type Mutation {                                 ' + sLineBreak +
    '  createAuthor(name: String): Author!           ' + sLineBreak +
    '  createBook(name: String, authorId: ID): Book! ' + sLineBreak +
    '  testUpdate(input: TestInput): TestOutput      ' + sLineBreak +
    '}                                               ' + sLineBreak +
    '                                                ' + sLineBreak +
    'type Query {                                    ' + sLineBreak +
    '  author(id: ID!): Author                       ' + sLineBreak +
    '  authors: [Author!]!                           ' + sLineBreak +
    '  book(id: ID!): Book                           ' + sLineBreak +
    '  books: [Book!]!                               ' + sLineBreak +
    '}';

procedure ResetCounter;
procedure InitBookshelfSchema(Schema: TSchemaDocument);

implementation

var
  Counter: Integer;
  BookList: TObjectList<TBook>;
  AuthorList: TObjectList<TAuthor>;

procedure InitBookshelfSchema(Schema: TSchemaDocument);
begin
  Schema.Bind('Query', TQuery);
  Schema.Bind('Mutation', TMutation);
  Schema.Bind('Author', TAuthor);
end;

procedure RegisterGlobalTypes(Schema: TSchemaDocument);
begin
  Schema.Add(TSchemaDateTimeType.Create);
end;

procedure ResetCounter;
begin
  Counter := 1;
end;

{ TBook }

constructor TBook.Create(AName: string; AAuthorID: Integer);
begin
  inherited Create;
  FName := AName;
  FAuthorID := AAuthorID;
  FId := Counter;
  Inc(Counter);
end;

function TBook.GetAuthor: TAuthor;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to AuthorList.Count - 1 do
    if AuthorList[I].Id = Self.FAuthorId then
    begin
      Result := AuthorList[I];
      Break;
    end;
end;

{ TMutation }

function TMutation.CreateAuthor(Name: string): TAuthor;
begin
  Result := TAuthor.Create(Name);
  try
    AuthorList.Add(Result)
  except
    Result.Free;
    raise;
  end;
end;

function TMutation.CreateBook(Name: string; AuthorID: Integer): TBook;
begin
  Result := TBook.Create(Name, AuthorID);
  try
    BookList.Add(Result)
  except
    Result.Free;
    raise;
  end;
end;

function TMutation.TestUpdate(Input: TTestInput): String;
begin
  // todo...some code will be added here later
end;

{ TAuthor }

constructor TAuthor.Create(AName: string);
begin
  inherited Create;
  FName := AName;
  FId := Counter;
  FCreatedOn := Now;
  Inc(Counter);
end;

function TAuthor.GetBooks: TArray<TBook>;
var
  I: Integer;
  Books: TList<TBook>;
begin
  Books := TList<TBook>.Create;
  try
    for I := 0 to BookList.Count - 1 do
      if BookList[I].FAuthorId = Self.Id then
        Books.Add(BookList[I]);

    Result := Books.ToArray;
  finally
    Books.Free;
  end;
end;

{ TQuery }

function TQuery.Author(id: Integer): TAuthor;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to AuthorList.Count - 1 do
    if AuthorList[I].Id = id then
    begin
      Result := AuthorList[I];
      Break;
    end;
end;

function TQuery.Authors: TArray<TAuthor>;
begin
  Result := AuthorList.ToArray;
end;

function TQuery.Book(id: Integer): TBook;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to BookList.Count - 1 do
    if BookList[I].Id = id then
    begin
      Result := TBook.Create(BookList[I].Name, BookList[I].Author.Id);
      TGraphQLExecuter.Current.AddOwnership(Result);
      Break;
    end;
end;

function TQuery.Books: TArray<TBook>;
begin
  Result := BookList.ToArray;
end;

{ TSchemaDateTimeype }

constructor TSchemaDateTimeType.Create;
begin
  inherited Create('DateTime');
end;

function TSchemaDateTimeType.ParseLiteral(Value: TASTValue): TValue;
var
  DateValue: TDateTime;
begin
  if Value is TASTStringValue then
  begin
    if not TBclUtils.TryISOToDateTime(TASTSTringValue(Value).Value, DateValue) then
      raise EGraphQLCoercingParseLiteral.Create(Self.DisplayName, Value);
    Result := TValue.From<TDateTime>(DateValue);
  end
  else
    raise EGraphQLCoercingParseLiteral.Create(Self.DisplayName, Value);
end;

function TSchemaDateTimeType.ParseValue(const Value: TValue): TValue;
begin
  if (Value.TypeInfo =  System.TypeInfo(TDateTime)) or
    (Value.TypeInfo =  System.TypeInfo(TDate)) or
    (Value.TypeInfo =  System.TypeInfo(TTime)) then
    Result := Value
  else
    raise EGraphQLCoercingParseValue.Create(Self.DisplayName, Value);
end;

function TSchemaDateTimeType.Serialize(const Value: TValue): TValue;
begin
  if Value.IsEmpty then
    Exit(TValue.Empty);

  if not Value.TryCast(TypeInfo(TDateTime), Result) then
    raise EGraphQLCoercingSerialize.Create(Self.DisplayName, Value);

  Result := TBclUtils.DateTimeToISO(Value.AsType<TDateTime>, True);
end;

{ TTestInput }

constructor TTestInput.Create(AName: String);
begin
  inherited Create;
  Self.MyName := AName;
end;

{ TTestOutput }

constructor TTestOutput.Create(AName: String);
begin
  inherited Create;
  Self.MyName := AName;
end;

initialization
  Counter := 1;
  BookList := TObjectList<TBook>.Create(True);
  AuthorList := TObjectList<TAuthor>.Create(True);
  TSchemaDocument.OnGlobalCreate := RegisterGlobalTypes;

finalization
  BookList.Free;
  AuthorList.Free;

end.
